/*
 * lesslog - PostgreSQL archive log size reduction utility
 *
 *    Copyright(C) 2007 Nippon Telegraph and Telephone Corporation
 */

/*
 * file.c
 *    Common I/O routine implementation used both in
 *    pgcompresslog/pgdecompresslog
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "postgres.h"
#include "access/xlog_internal.h"

#include "file.h"

/*
 * read_buff()
 *
 * Read the file data for specified lenth and return the size
 * actually read.
 *
 * Parameters:
 *    fd: File descriptor.
 *    buff: Buffer to read.
 *    len: Size to read.
 *
 * Note: If error occurd, exit(2) will be called here and will not return to the
 * caller in this case.
 */
int
read_buff(int fd, char *buff, size_t len)
{
	int ret;
	size_t read_len = 0;

	do
	{
		ret = read(fd, buff + read_len, len - read_len);
		if (ret < 0)
		{
			if (errno == EINTR)
				continue;
			fprintf(stderr, "failed to read : %s\n", strerror(errno));
			exit(1);
		}
		else if (ret == 0)
			break;
		read_len += ret;
	} while (read_len < len);

	return read_len;
}

/*
 * Write to the file and return length actually written.
 * The length to write should be specified as an argument.
 *
 * Parameter:
 *    fd: File descriptor.
 *    buff: Buffer to write.
 *    len: Size to write.
 *
 * Note: If an error occurs, exit(2) will be called in this function
 * and will not return to the caller in this case.
 */
void
write_buff(int fd, const char *buff, size_t len)
{
	int ret;
	int written_len = 0;

	do
	{
		ret = write(fd, buff + written_len, len - written_len);
		if (ret < 0)
		{
			if (errno == EINTR)
				continue;
			fprintf(stderr, "failed to write : %s\n", strerror(errno));
			exit(1);
		}
		written_len += ret;
	} while (written_len < len);
}

/*
 * Copy the contents of the file.
 *
 * Parameter:
 *    from_fd: File descriptor of the file to copy from.
 *    to_fd: File descriptor of the file to copy to.
 *
 * Note: If error occurs in this function, exit(2) will be called here
 * and will not return to the caller in this case.
 */
void
copy_file(int from_fd, int to_fd)
{
	int read_len = 0;
	char buff[8 * 1024];	/* 8KB buffer */

	while (1)
	{
		/* Read to the buffer. */
		read_len = read(from_fd, buff, sizeof(buff));
		if (read_len < 0)
		{
			if (errno == EINTR)
				continue;
			fprintf(stderr, "failed to read : %s\n", strerror(errno));
			exit(1);
		}
		else if (read_len == 0)
			break;
		/* Write all the buffer content. */
		write_buff(to_fd, buff, read_len);
	}

	return;
}

/*
 * Validate the record by comparing CRC value.
 *
 * CRC value will be calculated in the following order.
 *  - Logical Log
 *  - Full page write (if exists)
 *  - Record header (exluding the CRC area)
 *
 * Parameters:
 *    precord: Pointer to the target record.
 */
bool
is_valid_record(XLogRecord *precord)
{
	pg_crc32 crc;
	BkpBlock *pblk;
	int i;

	/* Calculate CRC for a logical log. */
	INIT_CRC32(crc);
	COMP_CRC32(crc, XLogRecGetData(precord), precord->xl_len);

	/*
	 * If full page writes exist, calculate CRC for each full page write.
	 */
	pblk = (BkpBlock *)((char *)XLogRecGetData(precord) + precord->xl_len);
	for (i = 0; i < XLR_MAX_BKP_BLOCKS; i++)
	{
		uint32 blen;

		if (!(precord->xl_info & XLR_SET_BKP_BLOCK(i)))
			continue;

		if (pblk->hole_offset + pblk->hole_length > BLCKSZ)
		{
			fprintf(stderr, "incorrect hole size in record.\n");
			return false;
		}

		blen = sizeof(BkpBlock) + BLCKSZ - pblk->hole_length;
		COMP_CRC32(crc, (char *)pblk, blen);
		pblk = (BkpBlock *)((char *)pblk + blen);
	}

	/* Calculate record header CRC value. */
	COMP_CRC32(crc, (char *)precord + sizeof(pg_crc32),
		SizeOfXLogRecord - sizeof(pg_crc32));

	/*
	 * Examine if the final CRC is the same as the value found in the
	 * record.
     */
	FIN_CRC32(crc);
	if (!EQ_CRC32(precord->xl_crc, crc))
	{
		fprintf(stderr, "incorrect resource manager data checksum.\n");
		return false;
	}

	return true;
}

/*
 * Calculate CRC of the record.
 */
pg_crc32
calc_crc(XLogRecord *record)
{
	int i;
	pg_crc32 crc;
	char *blk;

	INIT_CRC32((crc));

	/* logical log */
	COMP_CRC32((crc), XLogRecGetData((record)), record->xl_len);

	/* physical log */
	blk = (char *)XLogRecGetData(record) + record->xl_len;
	for (i = 0; i < XLR_MAX_BKP_BLOCKS; i++)
	{
		BkpBlock bkpb;
		uint32 blen;

		if (!(record->xl_info & XLR_SET_BKP_BLOCK(i)))
			continue;
		memcpy(&bkpb, blk, sizeof(BkpBlock));
		if (bkpb.hole_offset + bkpb.hole_length > BLCKSZ)
		{
			fprintf(stderr, "incorrect hole size in record\n");
			exit(1);
		}
		blen = sizeof(BkpBlock) + BLCKSZ - bkpb.hole_length;
		COMP_CRC32(crc, blk, blen);
		blk += blen;
	}

	/* record header expect xl_crc */
	COMP_CRC32((crc), (char *)record + sizeof(pg_crc32),
		SizeOfXLogRecord - sizeof(pg_crc32));

	FIN_CRC32((crc));

	return crc;
}
