/*
 * lesslog - PostgreSQL archive log size reduction utility
 *
 *    Copyright(C) 2007 Nippon Telegraph and Telephone Corporation
 */

/*
 * remove_handler.c
 *    replace backup block with incremental data while WAL archiving.
 *
 *    This file includes the code to extract incremental logs for each
 *    WAL record type.
 */
#include "postgres.h"
#include "access/gist_private.h"
#include "access/htup.h"
#include "access/nbtree.h"
#include "access/transam.h"
#include "access/xlog.h"

#include "remove.h"
#include "remove_handler.h"
#include "debug.h"

/*
 * Remove the backup blocks from XLOG record which doesn't need
 * incremental-format data.
 *
 * Note: callers must be sure that the backup blocks are removable.
 */
uint32
remove_common(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	return get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + record->xl_len, next_pos);
}

/*
 * Remove backup block from XLOG_HEAP_INSERT record.
 *
 * Inserted tuple needs to be extracted from bkpbk1.
 * It has the following fields:
 *  - xl_heap_header
 *  - inserted tuple data
 * the tuple is placed at the beginning of bkpbk1.
 *
 */
uint32
remove_heap_insert(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	xl_heap_insert *xlrec;
	BkpBlock *bkpb;
	PageHeader page;
	ItemId lp;
	HeapTupleHeader htup;
	xl_heap_header xlhdr;
	uint32 newlen;
	uint32 new_xl_len;
	uint32 noop_len;

	/* Extract inserted tuple from bkpbk1. */
	xlrec = (xl_heap_insert *)XLogRecGetData(record);
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len +
		sizeof(BkpBlock));
	lp = PageGetItemId(page, xlrec->target.tid.ip_posid);
	htup = (HeapTupleHeader)(((char *)page) + lp->lp_off - bkpb->hole_length);

	/* Fill in xl_heap_header. */
	xlhdr.t_infomask2 = htup->t_infomask2;
	xlhdr.t_infomask = htup->t_infomask;
	xlhdr.t_hoff = htup->t_hoff;

	/* the length of inserted tuple */
	newlen = lp->lp_len;

	/* Add the length of inserted tuple to xlog record. */
	new_xl_len = record->xl_len + SizeOfHeapHeader +
		(newlen - offsetof(HeapTupleHeaderData, t_bits));

	/* Calculate the size of removed space */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* We can place an XNOOP record, so don't remove backup block. */
	if (!noop_len)
		return 0;

	/* Copy inserted tuple to bkpbk1. */
	memcpy(bkpb, &xlhdr, SizeOfHeapHeader);
	memcpy((char *)bkpb + SizeOfHeapHeader,
		(char *)htup + offsetof(HeapTupleHeaderData, t_bits),
		newlen - offsetof(HeapTupleHeaderData, t_bits));

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return noop_len;
}

/*
 * Remove backup block from XLOG_HEAP_UPDATE or XLOG_HEAP_MOVE record.
 *
 * Updated tuple needs to be extracted from backup blocks. It has the
 * following fields:
 *  - xl_heap_header
 *  - new tuple data
 *
 * We don't care if xl_info has XLOG_HEAP_INIT_PAGE flag, just generate
 * xl_heap_header and new tuple data(null-bitmap, padding, oid, column-data)
 * from backup block.
 */
uint32
remove_heap_update(XLogRecord *record, const char *rec_pos, const char *next_pos,
	bool move)
{
	xl_heap_update *xlrec;
	BkpBlock *bkpb;
	bool	samepage;
	PageHeader page;
	ItemId lp;
	HeapTupleHeader htup;
	uint32 newlen;
	uint32 new_xl_len;
	uint32 noop_len;
	struct
	{
		xl_heap_header hdr;
		TransactionId tid1; /* xmax */
		TransactionId tid2; /* xmin */
	}	xlhdr;
	int hsize = SizeOfHeapHeader;

	/*
	 * Locate xlog data and backup block which contains the new tuple.
	 */
	xlrec = (xl_heap_update *)XLogRecGetData(record);
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)((char*)bkpb + sizeof(BkpBlock));
	samepage = (ItemPointerGetBlockNumber(&(xlrec->newtid)) ==
			ItemPointerGetBlockNumber(&(xlrec->target.tid)));

	/*
	 * If the record does not have bkpbk2, and tuples are not in the same page,
	 * we have enough information for recovery, so just remove backup block(s).
	 */
	if (!(record->xl_info & XLR_BKP_BLOCK_2) && !samepage)
	{
		return remove_common(record, rec_pos, next_pos);
	}

	/*
	 * If the record has bkpbk1, and tuples are NOT in the same page,
	 * skip bkpbk1 and locate bkpbk2.
	 */
	if ((record->xl_info & XLR_BKP_BLOCK_1) && !samepage)
	{
		bkpb = (BkpBlock *)((char *)bkpb + sizeof(BkpBlock) +
			(BLCKSZ - bkpb->hole_length));
		page = (PageHeader)((char *)bkpb + sizeof(BkpBlock));
	}

	/* Extract new tuple from backup block. */
	lp = PageGetItemId(page, xlrec->newtid.ip_posid);
	htup = (HeapTupleHeader)((char *)page + lp->lp_off - bkpb->hole_length);

	/* Fill in xl_heap_header. */
	xlhdr.hdr.t_infomask2 = htup->t_infomask2;
	xlhdr.hdr.t_infomask = htup->t_infomask;
	xlhdr.hdr.t_hoff = htup->t_hoff;

	/*
	 * The following code is taken from log_heap_update() in heapam.c
	 * (8.2.4). No essential changes are made in 8.3beta2.
	 */
	if (move)					/* remember xmax & xmin */
	{
		TransactionId xid[2];	/* xmin, xmax */

		if (htup->t_infomask & (HEAP_XMAX_INVALID | HEAP_IS_LOCKED))
			xid[0] = InvalidTransactionId;
		else
			xid[0] = HeapTupleHeaderGetXmax(htup);
		xid[1] = HeapTupleHeaderGetXmin(htup);
		memcpy((char *) &xlhdr + hsize, (char *)xid, 2 * sizeof(TransactionId));
		hsize += 2 * sizeof(TransactionId);
	}

	/* The length of new tuple data (bitmap [+ padding] [+ oid] + data). */
	newlen = lp->lp_len - offsetof(HeapTupleHeaderData, t_bits);

	/* Recalculate the length of new tuple. */
	new_xl_len = record->xl_len + hsize + newlen;

	/* Calculate the length of removed space. */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* We can't place an XNOOP record, so don't remove backup block */
	if (!noop_len)
		return 0;

	/*
	 * Copy new tuple into bkpbk1. It is possible we are at bkpbk2 now, so
	 * jump back to bkpbk1 before copying.
	 * (Koichi's suggestion)
	 * Copy the new tuple to bkpbk1.  Because we may be at bkpbk2
	 * now, jump back to bkpbk1 before copying.
	 */
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);

	memcpy(bkpb, &xlhdr, hsize);
	memcpy((char *)bkpb + hsize,
		(char *)htup + offsetof(HeapTupleHeaderData, t_bits), newlen);

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return noop_len;
}

/*
 * Remove backup block from XLOG_HEAP_INPLACE record.
 */
uint32
remove_heap_inplace(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	xl_heap_inplace *xlrec;
	BkpBlock *bkpb;
	PageHeader page;
	ItemId lp;
	HeapTupleHeader htup;

	uint32 newlen;
	uint32 new_xl_len;
	uint32 noop_len;

	/* Extract new tuple from bkpbk1 */
	xlrec = (xl_heap_inplace *)XLogRecGetData(record);
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len +
		sizeof(BkpBlock));

	lp = PageGetItemId(page, xlrec->target.tid.ip_posid);
	htup = (HeapTupleHeader)(((char *)page) + lp->lp_off - bkpb->hole_length);

	/* Calculate the size of new tuple.(not including header). */
	newlen = ItemIdGetLength(lp) - htup->t_hoff;

	/* Recalculate the length of XLOG record. */
	new_xl_len = record->xl_len + newlen;

	/* Calculate the length of removed space. */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* If We can't place an XNOOP record, don't remove backup block. */
	if (!noop_len)
		return 0;

	/* Copy new tuple to the bkpbk1 */
	memcpy((char *)bkpb, (char *)htup + htup->t_hoff, newlen);
#ifdef DEBUG
	fprintf(stderr, "%s(): tuple data (%d bytes) copied.\n", __func__, newlen);
#endif

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return noop_len;
}

/*
 * Determine if the tuple has already frozen by vacuum freeze (not
 * by the operation of the given WAL record).
 * If the tuple has already been frozen, return not zero.
 */
#define IS_FROZEN_TUPLE(tuple) \
( \
	( \
		(HeapTupleHeaderGetXvac((tuple)) == InvalidTransactionId) && \
		((tuple)->t_infomask & HEAP_MOVED_OFF) && \
		((tuple)->t_infomask & HEAP_XMIN_COMMITTED) \
	) ? 4 : \
	( \
		(HeapTupleHeaderGetXvac((tuple)) == FrozenTransactionId) && \
		((tuple)->t_infomask & HEAP_MOVED_IN) && \
		((tuple)->t_infomask & HEAP_XMIN_COMMITTED) \
	) ? 3 : \
	( \
		(HeapTupleHeaderGetXmax((tuple)) == InvalidTransactionId) && \
		(!((tuple)->t_infomask & HEAP_XMAX_IS_MULTI)) && \
		(!((tuple)->t_infomask & HEAP_XMAX_COMMITTED)) && \
		((tuple)->t_infomask & HEAP_XMAX_INVALID) \
	) ? 2 : \
	( \
		(HeapTupleHeaderGetXmin((tuple)) == FrozenTransactionId) && \
		((tuple)->t_infomask & HEAP_XMIN_COMMITTED) \
	) ? 1 : \
	0 \
)

/*
 * Remove backup block from XLOG_HEAP2_FREEZE record.
 */
uint32
remove_heap2_freeze(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	BkpBlock *bkpb;
	PageHeader page;
	OffsetNumber frozen[MaxOffsetNumber];
	int nfrozen;
	int offnum;

	uint32 new_xl_len;
	uint32 noop_len;

	/* Locate backup block. */
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len
			+ sizeof(BkpBlock));

	/* Reconstruct OffsetNumber array of frozen tuple from bkpbk1. */
	nfrozen = 0;
	for (offnum = FirstOffsetNumber;
		offnum <= PageGetMaxOffsetNumber(page);
		offnum = OffsetNumberNext(offnum))
	{
		ItemId lp = PageGetItemId(page, (OffsetNumber)offnum);

		if ItemIdIsUsed(lp)
		{
			HeapTupleHeader tuple =
				(HeapTupleHeader)((char *)page +
				lp->lp_off - bkpb->hole_length);

			if (IS_FROZEN_TUPLE(tuple))
				frozen[nfrozen++] = (OffsetNumber)offnum;
		}
	}

	/*
	 * NOTE: We don't need to assert if nfrozen == 0 or not here.
	 *
	 * Although there is at least one frozen tuple in the backup block of
	 * an XLOG_HEAP2_FREEZE record, but even there is no frozen tuple found,
	 * recovery process doesn't treat it as an error. So even nfrozen == 0,
	 * we still need to generate an XLOG for it.
	 *
	 * see also src/backend/access/heap/heapam.c
	 */

	/* Recompute the length of xlog record. */
	new_xl_len = record->xl_len + sizeof(OffsetNumber) * nfrozen;

	/* Calculate the length of removed space. */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* If we can't place an XNOOP record, don't remove backup block. */
	if (!noop_len)
		return 0;

	/* Copy offsetnum array of frozen tuple to the bkpbk1. */
	memcpy((char *)bkpb, frozen, nfrozen * sizeof(OffsetNumber));
#ifdef DEBUG
	fprintf(stderr, "%s(): frozen item array (%d item, %d bytes) copied.\n", __func__, nfrozen, sizeof(OffsetNumber) * nfrozen);
#endif

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return noop_len;
}

/*
 * Remove backup block from XLOG_BTREE_INSERT_* record.
 *   - XLOG_BTREE_INSERT_LEAF
 *   - XLOG_BTREE_INSERT_UPPER
 *   - XLOG_BTREE_INSERT_META
 */
uint32
remove_btree_insert(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	xl_btree_insert *xlrec;
	BkpBlock *bkpb;
	PageHeader page;
	ItemId iid;
	IndexTuple ituple;

	uint32 tlen;
	uint32 new_xl_len;
	uint32 noop_len;

	/* Extract index tuple from bkpbk1. */
	xlrec = (xl_btree_insert *)XLogRecGetData(record);
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len +
		sizeof(BkpBlock));

	iid = PageGetItemId(page, xlrec->target.tid.ip_posid);
	ituple = (IndexTuple)(((char *)page) + iid->lp_off - bkpb->hole_length);

	/*
	 * Here we don't care if the tuple was stripped down to just
	 * the tuple header, because the tuple has the actual length
	 * inside.  (Read comments in _bt_pgaddtup)
	 */
	tlen = IndexTupleSize(ituple);

	/* Calculate the length of XLOG record(including inserted tuple). */
	new_xl_len = record->xl_len + tlen;

	/* Calculate the length of removed space. */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* If we can't place an XNOOP record, don't remove backup block. */
	if (!noop_len)
		return 0;

	/* Copy index tuple to the beginning of bkpblock 1. */
	memcpy((char *)bkpb, (char *)ituple, tlen);

#ifdef DEBUG
	fprintf(stderr, "%s(): index tuple( offset=%u, t_ctid=%u, %d byte)copied.\n", __func__, xlrec->target.tid.ip_posid, ituple->t_tid.ip_posid, tlen);
#endif

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return noop_len;
}

/*
 * Remove backup block from XLOG_BTREE_SPLIT_L_* record.
 *   - XLOG_BTREE_SPLIT_L
 *   - XLOG_BTREE_SPLIT_L_ROOT
 */
uint32
remove_btree_split_left(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	xl_btree_split *xlrec;
	BkpBlock *bkpb;
	PageHeader page;

	ItemId iid;
	IndexTuple ituple;			/* new index tuple */
	IndexTuple htuple = NULL;	/* HIKEY index tuple */
	uint32 itlen;
	uint32 htlen = 0;
	char ituple_buff[BLCKSZ];	/* buffer for new index tuple */
	char htuple_buff[BLCKSZ];	/* buffer for HIKEY index tuple */

	OffsetNumber newitemoff;
	char *datapos;
	uint32 datalen;
	uint32 new_xl_len = record->xl_len;
	uint32 noop_len;

	/* If BkpBlock_1 does not exist, we have enough information already. */
	if (!(record->xl_info & XLR_BKP_BLOCK_1))
		return remove_common(record, rec_pos, next_pos);

	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len +
		sizeof(BkpBlock));

	xlrec = (xl_btree_split *)XLogRecGetData(record);
	datapos = (char *) xlrec + SizeOfBtreeSplit;
	datalen = record->xl_len - SizeOfBtreeSplit;

	/* Find HIKEY tuple in the backup block if split page was not a leaf. */
	if (xlrec->level > 0)
	{
		/* skip downlink */
		datapos += sizeof(BlockIdData);
		datalen -= sizeof(BlockIdData);

		/* Calculate the length of XLOG record(including HIKEY index tuple) */
		iid = PageGetItemId(page, P_HIKEY);
		htuple = (IndexTuple)(((char *)page) + iid->lp_off - bkpb->hole_length);
		htlen = MAXALIGN(IndexTupleSize(htuple));
		new_xl_len += htlen;
	}

	/* Extract offset of new index tuple. */
	memcpy(&newitemoff, datapos, sizeof(OffsetNumber));
	datapos += sizeof(OffsetNumber);
	datalen -= sizeof(OffsetNumber);

	iid = PageGetItemId(page, newitemoff);
	ituple = (IndexTuple)(((char *)page) + iid->lp_off - bkpb->hole_length);

	/* Calculate the length of XLOG record(including new index tuple) */
	itlen = MAXALIGN(IndexTupleSize(ituple));
	new_xl_len += itlen;

	/* Calculate the length of removed space */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* If we can't place an XNOOP record, don't remove backup blocks. */
	if (!noop_len)
		return 0;

	/* Prevent overwrite with right-sibling data */
	memcpy(ituple_buff, ituple, IndexTupleSize(ituple));
	if (xlrec->level > 0)
		memcpy(htuple_buff, htuple, IndexTupleSize(htuple));

	/* Move right sibling data backward to make up space for incremental log. */
	memmove(datapos + itlen + htlen, datapos, datalen);

	/* Insert new tuple (and HIKEY tuple if necessary) into XLOG record. */
	if (xlrec->level > 0)
	{
		datapos -= sizeof(OffsetNumber);
		memcpy(datapos, htuple_buff, htlen);
		memcpy(datapos + htlen, &newitemoff, sizeof(OffsetNumber));
		datapos += htlen + sizeof(OffsetNumber);
	}
	memcpy(datapos, ituple_buff, itlen);
		
#ifdef DEBUG
	fprintf(stderr, "%s(): new index tuple( t_ctid=%u, %d byte)insert. \n",
		__func__, ((IndexTuple)(ituple_buff))->t_tid.ip_blkid, itlen);
#endif

	/* Update length info in XLOG header. */
	record->xl_len = new_xl_len;

	return  noop_len;
}

/*
 * Remove backup block from XLOG_GIST_NEW_ROOT record.
 */
uint32
remove_gist_new_root(XLogRecord *record, const char *rec_pos, const char *next_pos)
{
	gistxlogPageUpdate *xlrec;
	BkpBlock *bkpb;
	PageHeader page;

	uint32 new_xl_len;
	uint32 noop_len;
	int offnum,
		tsize = 0,
		maxoff,
		nused = 0,
		newlen;
	char tuple_buff[BLCKSZ];
	char *pos;

	/* Locate bkpblock. */
	xlrec = (gistxlogPageUpdate *)XLogRecGetData(record);
	bkpb = (BkpBlock *)(XLogRecGetData(record) + record->xl_len);
	page = (PageHeader)(XLogRecGetData(record) + record->xl_len +
		sizeof(BkpBlock));

	/*
	 * Scan the backup block and calculate the total size of used tuples.
	 * They will be added to the existing incremental log.
	 */
	maxoff = PageGetMaxOffsetNumber(page);
	pos = tuple_buff;
	newlen = 0;
	for (offnum = FirstOffsetNumber; offnum <= maxoff;
		offnum = OffsetNumberNext(offnum))
	{
		ItemId lp = PageGetItemId(page, (OffsetNumber)offnum);
		IndexTuple t;

		if (!ItemIdIsUsed(lp))
			continue;

		t = (IndexTuple) ((char *)page + lp->lp_off - bkpb->hole_length);
		tsize = IndexTupleSize(t);
		memcpy(pos, t, tsize);
		newlen += tsize;
		pos += tsize;
		nused++;
	}

	/* the size of extracted tuples */
	newlen = tsize * nused;

	/*
	 * Calculate the length of XLOG record(including index tuples in root page).
	 */
	new_xl_len = record->xl_len + newlen;

	/* Calculate the length of removed space. */
	noop_len = get_noop_len(rec_pos, record->xl_tot_len,
		SizeOfXLogRecord + new_xl_len, next_pos);

	/* If we can't place an XNOOP record, don't remove backup blocks. */
	if (!noop_len)
		return 0;

	/* Copy extracted tuples found in the root page into bkpbk1. */
	memcpy((char *)bkpb, tuple_buff, newlen);
#ifdef DEBUG
	fprintf(stderr, "%s(): index tuples( size=%d, num=%d, total len=%d )extracted from root page. \n", __func__, tsize, nused, newlen);
#endif

	/* Update length info in XLOG header */
	record->xl_len = new_xl_len;

	return  noop_len;
}
